
#include <stdio.h>
#include <z80regs.h>

/* #option MATHLIB */
#option INLIB

/*********************************************************************************************************/

#define VKEY_LEFTARROW  8
#define VKEY_RIGHTARROW 9
#define VKEY_DOWNARROW  10
#define VKEY_UPARROW    11
#define VKEY_ENTER      13
#define VKEY_ESC        128

#define VKEY_SHIFT_LEFTARROW  24
#define VKEY_SHIFT_RIGHTARROW 25

#define VKEY_CLEAR_0 176
#define VKEY_CLEAR_1 177
#define VKEY_CLEAR_2 178
#define VKEY_CLEAR_3 179
#define VKEY_CLEAR_4 180
#define VKEY_CLEAR_5 181
#define VKEY_CLEAR_6 182
#define VKEY_CLEAR_7 183
#define VKEY_CLEAR_8 184
#define VKEY_CLEAR_9 185

#define VKEY_CLEAR_C 195
#define VKEY_CLEAR_D 196
#define VKEY_CLEAR_E 197
#define VKEY_CLEAR_H 200
#define VKEY_CLEAR_I 201
#define VKEY_CLEAR_S 211
#define VKEY_CLEAR_X 216

#define VKEY_F1 129
#define VKEY_F2 130
#define VKEY_F3 131

/*********************************************************************************************************/

#define HEADER_SIZE  20
#define GAURD_SIZE   32
#define LINE_LEN     81
#define HISTORY_SIZE 5
#define MEM_BUFFER_SIZE (HEADER_SIZE + (LINE_LEN * HISTORY_SIZE) + GAURD_SIZE)

typedef unsigned char byte;
typedef unsigned short int word;

typedef struct {
	char szBuf[LINE_LEN];
} CmdType;

CmdType g_ctHstCmd[HISTORY_SIZE];
char    g_szCmdBuf[LINE_LEN];

char g_bInsert;
int  g_nCurHst;
int  g_nCurY;
int  g_nCurX;
int  g_nCurLen;

word g_wPerBuf; /* address of the persistent buffer as allocated by the first time DoskKey is run */

/*********************************************************************************************************/
DebugMsg(pszMsg)
char* pszMsg;
{
	puts(pszMsg);
	getchar();
}

/*********************************************************************************************************/
char GetScrChar(x, y)
int x;
int y;
{
	union REGS reg;

	/****************************************************************************/
	/* @VDCTL = SVC-15 */

	reg.byteregs.a = 15;
	reg.byteregs.b = 1;
	reg.byteregs.h = y;
	reg.byteregs.l = x;
	call(15, &reg);

	return reg.byteregs.a;
}

/*********************************************************************************************************/
SetCursorChar(c)
unsigned char c;
{
	union REGS reg;

	/****************************************************************************/
	/* @VDCTL = SVC-15 */

	reg.byteregs.a = 15;
	reg.byteregs.b = 8;
	reg.byteregs.c = c;
	call(15, &reg);
}

/*********************************************************************************************************/
/* returns: 0 on failure; the low memory address on success; */
word GetLowMem()
{
	union REGS reg;

	/****************************************************************************/
	/* @HIGH$ = SVC-100 */

	reg.byteregs.a  = 100;
	reg.byteregs.b  = 1;
	reg.wordregs.hl = 0;
	call(100, &reg);

/*	printf("GetLowMem(), F %d\n", reg.byteregs.f); /* */

	if (reg.byteregs.f & 0x40) /* then success */
	{
		return reg.wordregs.hl;
	}

	return 0;
}

/*********************************************************************************************************/
/* returns: 0 on failure; the high memory address on success; */
word GetHighMem()
{
	union REGS reg;

	/****************************************************************************/
	/* @HIGH$ = SVC-100 */

	reg.byteregs.a  = 100;
	reg.byteregs.b  = 0;
	reg.wordregs.hl = 0;
	call(100, &reg);

/*	printf("GetHighMem(), F %d\n", reg.byteregs.f); /* */

	if (reg.byteregs.f & 0x40) /* then success */
	{
		return reg.wordregs.hl;
	}

	return 0;
}

/*********************************************************************************************************/
/* returns: 1 on success; 0 on failure; */
char SetLowMem(nAddr)
word nAddr;
{
	union REGS reg;

	/****************************************************************************/
	/* @HIGH$ = SVC-100 */

	reg.byteregs.a  = 100;
	reg.byteregs.b  = 1;
	reg.wordregs.hl = nAddr;
	call(100, &reg);

/*	printf("SetLowMem(), F %d\n", reg.byteregs.f);  /* */

	if (reg.byteregs.f & 0x40) /* then success */
	{
		return 1;
	}

	return 0;
}

/*********************************************************************************************************/
/* returns: 1 on success; 0 on failure; */
char SetHighMem(nAddr)
word nAddr;
{
	union REGS reg;

	/****************************************************************************/
	/* @HIGH$ = SVC-100 */

	reg.byteregs.a  = 100;
	reg.byteregs.b  = 0;
	reg.wordregs.hl = nAddr;
	call(100, &reg);

/*	printf("SetHighMem(), F %d\n", reg.byteregs.f);  /* */

	if (reg.byteregs.f & 0x40) /* then success */
	{
		return 1;
	}

	return 0;
}

/*********************************************************************************************************/
/* returns: 0 on failure; the address of the memory module; */
word GetMod(pszName)
char* pszName;
{
	union REGS reg;

	/****************************************************************************/
	/* @GTMOD = SVC-83 */

	reg.byteregs.a = 83;
	reg.wordregs.de = pszName;
	call(83, &reg);

/*	printf("GetMod(), F %d\n", reg.byteregs.f);  /* */

	if (reg.byteregs.f & 0x40) /* if Zero flag is set then the module was found */
	{
		return reg.wordregs.hl;
	}
	else
	{
		return 0;
	}
}

/*********************************************************************************************************/
/* returns: 0 on failure; the address of the memory module; */
word LocateMod(pszName)
char* pszName;
{
	char* psz;
	word  wHigh = GetHighMem();

	psz = wHigh;

	while (psz < 0xFFFF)
	{
		if ((*psz == *pszName) && (memcmp(psz, pszName, 6) == 0) && (*(psz + 10) == 0xC9))
		{
			return psz;
		}

		++psz;
	}

	return 0;
}

/*********************************************************************************************************/
Install()
{
	byte byBuf[MEM_BUFFER_SIZE];
	word wHigh    = GetHighMem();
	word wNewHigh = wHigh - MEM_BUFFER_SIZE;
	int  i;

	g_wPerBuf = GetMod("DOSKEY");

	if (g_wPerBuf != 0)
	{
		return;
	}

	if (SetHighMem(wNewHigh) == 0)
	{
		return;
	}

	memset(byBuf, 0, sizeof(byBuf));

	/* JR e instruction */
	byBuf[0] = 0x18; /* JR opcode */
	byBuf[1] = 0x0D; /* relative address (15-2) */

	/* highest memory address used by this memory module */
	byBuf[2] = wHigh & 0xFF; /* LSB first */
	byBuf[3] = (wHigh >> 8);

	/* 6 characters for the memory module name */
	byBuf[4] = 6;

	/* set memory module name to "DOSKEY" */
	byBuf[5]  = 'D';
	byBuf[6]  = 'O';
	byBuf[7]  = 'S';
	byBuf[8]  = 'K';
	byBuf[9]  = 'E';
	byBuf[10] = 'Y';

	/* Device Control Block for memory module */
	byBuf[11] = 0;
	byBuf[12] = 0;

	/* reserved word */
	byBuf[13] = 0;
	byBuf[14] = 0;

	byBuf[15] = 0xC9; /* ret opcode */

	memcpy(wNewHigh+1, byBuf, sizeof(byBuf)-GAURD_SIZE);

	g_wPerBuf = GetMod("DOSKEY");
}

/*********************************************************************************************************/
DispCmd(szBuf)
char* szBuf;
{
	int i, nLen;

	cursor(0, g_nCurY);

	nLen = strlen(szBuf);

	if (nLen > g_nCurLen)
	{
		g_nCurLen = nLen;
	}

	for (i = 0; i < g_nCurLen; ++i)
	{
		if (i < nLen)
		{
			putchar(szBuf[i]);
		}
		else
		{
			putchar(' ');
		}
	}

	g_nCurX   = nLen;
	g_nCurLen = nLen;

	cursor(g_nCurX, g_nCurY);
}

/*********************************************************************************************************/
SelPrev()
{
	int i;

	if (g_nCurHst >= (HISTORY_SIZE-1))
	{
		return;
	}

	++g_nCurHst;

	if (g_ctHstCmd[g_nCurHst].szBuf[0] == 0)
	{
		--g_nCurHst;
		return;
	}

	strcpy(g_szCmdBuf, g_ctHstCmd[g_nCurHst].szBuf);
	DispCmd(g_szCmdBuf);
}

/*********************************************************************************************************/
SelNext()
{
	int i;

	if (g_nCurHst <= 0)
	{
		return;
	}

	--g_nCurHst;

	if (g_ctHstCmd[g_nCurHst].szBuf[0] == 0)
	{
		return;
	}

	strcpy(g_szCmdBuf, g_ctHstCmd[g_nCurHst].szBuf);
	DispCmd(g_szCmdBuf);
}

/*********************************************************************************************************/
readline(f, szLine, nMaxLen)
FILE* f;
char* szLine;
int   nMaxLen;
{
	char c;
	int  i;

	i = 0;
	c = fgetc(f);

	while (!feof(f) && (c != '\n') && (i < nMaxLen))
	{
		if ((c != '\n') && (c != '\r'))
		{
			szLine[i] = c;
		}

		++i;
		c = fgetc(f);
	}

	szLine[i] = 0;
}

/*********************************************************************************************************/
int IsEmptyString(psz)
char* psz;
{
	while (*psz != 0)
	{
		if (*psz != ' ')
		{
			return FALSE;
		}

		++psz;
	}

	return TRUE;
}

/*********************************************************************************************************/
char* getline(psz, szLine, nMaxLen)
char* psz;
char* szLine;
int   nMaxLen;
{
	char c;
	int  i;

	i = 0;
	c = *psz;
	++psz;

	while ((c != 0) && (c != '\n') && (i < nMaxLen))
	{
		if ((c != '\n') && (c != '\r'))
		{
			szLine[i] = c;
		}

		++i;
		c = *psz;
		++psz;
	}

	szLine[i] = 0;

	return psz;
}

/*********************************************************************************************************/
LoadHistory()
{
	char* psz;
	FILE* f;
	int   i;

	for (i = 0; i < HISTORY_SIZE; ++i)
	{
		g_ctHstCmd[i].szBuf[0] = 0;
	}

	if (g_wPerBuf == 0)
	{
		f = fopen("doskey/txt:7", "r"); /* try using ram drive first */

		if (f == NULL)
		{
			f = fopen("doskey/txt:0", "r"); /* if no drive 7 then try drive 0 */

			if (f == NULL)
			{
				return;
			}
		}

		i = 0;

		while (!feof(f) && (i < HISTORY_SIZE))
		{
			readline(f, g_ctHstCmd[i].szBuf, sizeof(g_ctHstCmd[i].szBuf)-1);

			if (IsEmptyString(g_ctHstCmd[i].szBuf) == FALSE)
			{
				++i;
			}
		}

		fclose(f);
	}
	else /* get from high memory */
	{
		i = 0;

		psz = g_wPerBuf + HEADER_SIZE;

		while ((*psz != 0) && (i < HISTORY_SIZE))
		{
			psz = getline(psz, g_ctHstCmd[i].szBuf, sizeof(g_ctHstCmd[i].szBuf) - 1);

			if (IsEmptyString(g_ctHstCmd[i].szBuf) == FALSE)
			{
				++i;
			}
		}
	}

	g_nCurHst = -1;
}

/*********************************************************************************************************/
SaveHistory()
{
	char* psz;
	FILE* f;
	int   i;

	if (g_wPerBuf == 0)
	{
		f = fopen("doskey/txt:7", "w");

		if (f == NULL)
		{
			f = fopen("doskey/txt:0", "w");

			if (f == NULL)
			{
				puts("Unable to open doskey/txt:0 to save history.");
				return;
			}
		}

		fputs(g_szCmdBuf, f);
		fputs("\n", f);

		for (i = 0; i < HISTORY_SIZE; ++i)
		{
			if ((i != g_nCurHst) || (strcmp(g_szCmdBuf, g_ctHstCmd[i].szBuf) != 0)) /* don't save an unmodified version of the current history selection */
			{
				if (IsEmptyString(g_ctHstCmd[i].szBuf) == FALSE) /* don't save empty strings */
				{
					fputs(g_ctHstCmd[i].szBuf, f);
					fputs("\n", f);
				}
			}
		}

		fclose(f);
	}
	else
	{
		psz = g_wPerBuf + HEADER_SIZE;

		strcpy(psz, g_szCmdBuf);
		strcat(psz, "\n");

		for (i = 0; i < HISTORY_SIZE; ++i)
		{
			if ((i != g_nCurHst) || (strcmp(g_szCmdBuf, g_ctHstCmd[i].szBuf) != 0)) /* don't save an unmodified version of the current history selection */
			{
				if (IsEmptyString(g_ctHstCmd[i].szBuf) == FALSE) /* don't save empty strings */
				{
					strcat(psz, g_ctHstCmd[i].szBuf);
					strcat(psz, "\n");
				}
			}
		}
	}
}

/*********************************************************************************************************/
Execute(pszCmd)
char* pszCmd;
{
	int i;

	/* trim leading blanks */
	while (*pszCmd == ' ')
	{
		++pszCmd;
	}

	/* trim trailing blanks */
	i = strlen(pszCmd);

	while ((i > 0) && (pszCmd[i] == ' '))
	{
		--i;
	}

	pszCmd[i+1] = 0;

	SetCursorChar('_');

	cursor(0, g_nCurY);
	puts(pszCmd);
	cmdi(pszCmd);
}

/*********************************************************************************************************/
DelChar(szBuf, nPos)
char* szBuf;
int nPos;
{
	while (szBuf[nPos] != 0)
	{
		szBuf[nPos] = szBuf[nPos+1];

		if (szBuf[nPos] != 0)
		{
			putchar(szBuf[nPos]);
		}
		else
		{
			putchar(' ');
		}

		++nPos;
	}
}

/*********************************************************************************************************/
InsertChar(szBuf, nPos, c)
char* szBuf;
int   nPos;
char  c;
{
	int i, nLen;

	nLen = strlen(szBuf);
	i = nLen;

	szBuf[i+1] = 0;

	--i;

	while (i >= nPos)
	{
		szBuf[i+1] = szBuf[i];
		--i;
	}

	szBuf[nPos] = c;
}

/*********************************************************************************************************/
ShowHelp()
{
	cursor(0, 0);
	puts("\x1f");               /* erase to end of screen */
	cursor(0, 0);
	puts("Doskey version 1.0");
	puts(" ");
	puts("Usage: doskey command parm1 parm2 .....");
	puts(" ");
	puts("If command and parms are not specified then you can navigate");
	puts("though previous commands using the following key strokes.");
	puts(" ");
	puts("Enter       - Execute the command line.");
	puts("Break       - Exit without executing the command line.");
	puts("Clear + C   - Erase entire command line.");
	puts("Clear + I   - Toggle between insert and overtype modes.");
	puts("Clear + D   - Delete character at cursor position.");
	puts("Clear + E   - Erase from cursor to end of line.");
	puts("Left Arrow  - Move cursor one position to the left (non destructive).");
	puts("Right Arrow - Move the cursor one poition to the right (non destructive).");
	puts("Up Arrow    - Replace command with previous.");
	puts("Down Arrow  - Replace command with next.");
	puts("Shift Left Arrow  - Move cursor to the start of the line.");
	puts("Shift Right Arrow - Move the cursor to the end of the line.");
	puts(" ");

	g_nCurY = (g_nCurX = curpos()) >> 8;    /* get curpos; calc row */
	g_nCurX = g_nCurX & 0xFF;
}

/*********************************************************************************************************/
ClrLine(nLen)
int nLen;
{
	int i;

/*	for (i = g_nCurX; i < nLen; ++i)
	{
		putchar(' ');
	}
*/
	putchar(0x1E);
	cursor(g_nCurX, g_nCurY);
}

/*********************************************************************************************************/
/*
* ENTER       - Exit while executing the command line.
* Clear+C     - Clear command line.
* Cleat+I     - Toggle between insert and overtype modes.
* Clear+D     - Delete character under cursor.
* Clear+E     - Erase from cursor position to end of line.
* Break       - Exit without executing the command line.
* Left Arrow  - Move cursor one position to the left (non destructive).
* Right Arror - Move the cursor one poition to the right (non destructive).
* Up Arrow    - Replace command line text with the one entered before tyhe current selection.
* Down Arrow  - Replace command line text with the one entered after the current selection.
* Shift Left Arrow  - Move cursor to the start of the line.
* Shift Right Arrow - Move the cursor to the end of the line.
*/
ServiceKey(c)
char c;
{
	switch (c)
	{
		case VKEY_ENTER:			/* execute command line */
			SaveHistory();
			Execute(g_szCmdBuf);
			break;

		case VKEY_LEFTARROW:		/* move cursor one position to the left */
			if (g_nCurX <= 0)
			{
				break;
			}

			--g_nCurX;
			cursor(g_nCurX, g_nCurY);
			break;

		case VKEY_RIGHTARROW:		/* move cursor one position to the right */
			if (g_nCurX >= g_nCurLen)
			{
				break;
			}

			++g_nCurX;
			cursor(g_nCurX, g_nCurY);
			break;

		case VKEY_UPARROW:			/* select previous command in history buffer */
			SelPrev();
			break;

		case VKEY_DOWNARROW:		/* select next command in history buffer */
			SelNext();
			break;

		case VKEY_SHIFT_LEFTARROW:	/* move cursor to start of line */
			g_nCurX = 0;
			cursor(g_nCurX, g_nCurY);
			break;

		case VKEY_SHIFT_RIGHTARROW:	/* move cursor to end of line */
			g_nCurX = g_nCurLen;
			cursor(g_nCurX, g_nCurY);
			break;

		case VKEY_CLEAR_E:			/* erase to end of line */
			ClrLine(g_nCurLen);
			g_szCmdBuf[g_nCurX] = 0;
			g_nCurLen = g_nCurX;
			break;

		case VKEY_F1:				/* display help text */
			ShowHelp();
			DispCmd(g_szCmdBuf);
			break;

		case VKEY_CLEAR_I:			/* toggle insert/overtype mode */
		case VKEY_CLEAR_1:
			if (g_bInsert)
			{
				g_bInsert = FALSE;
				SetCursorChar(191);
			}
			else
			{
				g_bInsert = TRUE;
				SetCursorChar('_');
			}

			break;

		case VKEY_CLEAR_D:			/* delete character at cursor position */
		case VKEY_CLEAR_3:
			DelChar(g_szCmdBuf, g_nCurX);
			cursor(g_nCurX, g_nCurY);
			g_nCurLen = strlen(g_szCmdBuf);
			break;

		case VKEY_CLEAR_C:			/* delete entire line */
			cursor(0, g_nCurY);
			putchar('\x1f');        /* erase to end of screen */
			cursor(0, g_nCurY);
			g_szCmdBuf[0] = 0;
			g_nCurX   = 0;
			g_nCurLen = 0;
			break;

		default:
			if (g_nCurX >= 78)
			{
				break;
			}

			if (g_bInsert)
			{
				if (g_nCurLen >= 78)
				{
					break;
				}

				InsertChar(g_szCmdBuf, g_nCurX, c);
				printf("%s", g_szCmdBuf+g_nCurX);
				cursor(g_nCurX+1, g_nCurY);
				++g_nCurLen;
			}
			else
			{
				putchar(c);
				g_szCmdBuf[g_nCurX] = c;
			}

			++g_nCurX;

			if (g_nCurX > g_nCurLen)
			{
				g_nCurLen = g_nCurX;
			}

			g_szCmdBuf[g_nCurLen] = 0;
			break;
	}
}

/*********************************************************************************************************/
GetScrLine(szBuf, y)
char* szBuf;
int y;
{
	int x;

	for (x = 0; x < 78; ++x)
	{
		szBuf[x] = GetScrChar(x, y);
	}

	szBuf[x] = 0;
}

/*********************************************************************************************************/
TrimTrailingSpaces(szBuf)
char *szBuf;
{
	int i = strlen(szBuf) - 1;

	while ((i > 0) && (szBuf[i] == ' '))
	{
		szBuf[i] = 0;
		--i;
	}
}

/**************************************************************************************************************/
/* scans the display "history" for the "DOS Ready" test preceeding the one that launched this program.        */
/* then copies the text of the line following it to the edit buffer and presents it as an option to the user. */
GetLastCmd(szBuf)
char* szBuf;
{
	int x, y, nCurY, i;
	
	y = nCurY = curpos() >> 8;    /* get curpos; calc row */

	/* scan back two occurances of "DOS Ready" */
	i = 0;

	while ((i < 2) && (y > 0))
	{
		--y;
		GetScrLine(szBuf, y);

		if (strstr(szBuf, "DOS Ready") != 0)
		{
			++i;
		}
	}

	if (y == 0)
	{
		szBuf[0] = 0;
		return;
	}

	/* last command should be the line below "DOS Ready" */
	++y;
	GetScrLine(szBuf, y);
	TrimTrailingSpaces(szBuf);

	cursor(0, nCurY);
}

/*********************************************************************************************************/
main(argc, argv)
int   argc;
char* argv[];
{
	char c, i, bCmdSpecified;

	Install();
	LoadHistory();

	bCmdSpecified = 0;
	g_szCmdBuf[0] = 0;

	if ((argv[0][0] == '*') && (argv[0][1] != 0))
	{
		strcpy(g_szCmdBuf, argv[0]+1);
		bCmdSpecified = 1;
	}

	if (argc > 1)
	{
		if (bCmdSpecified)
		{
			strcat(g_szCmdBuf, " ");
		}

		strcat(g_szCmdBuf, argv[1]);

		for (i = 2; i < argc; ++i)
		{
			strcat(g_szCmdBuf, " ");
			strcat(g_szCmdBuf, argv[i]);
		}

		bCmdSpecified = 1;
	}

	if (bCmdSpecified)
	{
		SaveHistory();
		Execute(g_szCmdBuf);
	}

	GetLastCmd(g_szCmdBuf);

	g_nCurLen = strlen(g_szCmdBuf);

	g_bInsert = FALSE;
	g_nCurX   = 0;
	g_nCurY   = curpos() >> 8;    /* get curpos; calc row */

	puts("F1 for help");

	if (g_szCmdBuf[0] != 0)
	{
		DispCmd(g_szCmdBuf);
	}

	c = 0;

	while (c != VKEY_ESC)
	{
		c = getchar();
		ServiceKey(c);
/*		cursor(0, 0);
		printf("g_nCurX %d, g_nCurLen %d, c %d     ", g_nCurX, g_nCurLen, c);
		cursor(g_nCurX, g_nCurY); /* */
	}

	return 0;
}
